#!/usr/bin/env python

import math
import types
import re
from struct import *

# - binary pack reader
def ReadStruct(fp, format):
    try:
        length = calcsize(format)
        dat = fp.read(length)
        p = unpack(format,dat)
        if type(p) == types.TupleType:
            q = p[0]
        else:
            q = p
        return q            
    except:
        return 0

def StringRemoveNull(st):
    for pos , w in enumerate(st):
        if w == "\x00":
            return st[:pos]
    return st

def DecodeSjis(string):
    return string.decode('shift_jis','ignore') 

def ReadStringSjis(fp,length):
    return  DecodeSjis(StringRemoveNull(ReadStruct(fp,str(length)+"s")))

#
# --- 3D matrics
#
def matmul(a, b):
    c = [0,0,0,0,0,0,0,0,0]
    c[0] = a[0]*b[0]+a[1]*b[3]+a[2]*b[6]
    c[1] = a[0]*b[1]+a[1]*b[4]+a[2]*b[7]
    c[2] = a[0]*b[2]+a[1]*b[5]+a[2]*b[8]
    c[3] = a[3]*b[0]+a[4]*b[3]+a[5]*b[6]
    c[4] = a[3]*b[1]+a[4]*b[4]+a[5]*b[7]
    c[5] = a[3]*b[2]+a[4]*b[5]+a[5]*b[8]
    c[6] = a[6]*b[0]+a[7]*b[3]+a[8]*b[6]
    c[7] = a[6]*b[1]+a[7]*b[4]+a[8]*b[7]
    c[8] = a[6]*b[2]+a[7]*b[5]+a[8]*b[8]
    return c

def matrevrot(a):
    b = [0,0,0,0,0,0,0,0,0]
    b[0] = a[0]
    b[1] = a[3]
    b[2] = a[6]
    b[3] = a[1]
    b[4] = a[4]
    b[5] = a[7]
    b[6] = a[2]
    b[7] = a[5]
    b[8] = a[8]
    return b

def matrotx(r):
    m = [1.0, 0.0, 0.0,
         0.0, math.cos(r), -math.sin(r),
         0.0, math.sin(r), math.cos(r)]
    return m

def matroty(r):
    m = [math.cos(r), 0.0, math.sin(r),
         0.0, 1.0, 0.0,
         -math.sin(r), 0.0, math.cos(r)]
    return m

def matrotz(r):
    m = [math.cos(r), -math.sin(r), 0.0,
         math.sin(r), math.cos(r), 0.0,
         0.0, 0.0, 1.0]
    return m

def matrixToZXY(mat):
    rot_x = 0.0
    rot_y = 0.0
    rot_z = 0.0
    rot_x = -math.asin(mat[7])
    if (math.cos(rot_x) != 0.0):
        rot_y = math.atan2(mat[6], mat[8])
        alpha = mat[1]/math.cos(rot_x)
        if alpha > 1.0:
            alpha = 1.0
        if alpha < -1.0:
            alpha = -1.0
        rot_z = math.asin(alpha)
        if (mat[4] < 0.0):
            rot_z = math.pi - rot_z
    else:
        rot_y = math.atan2(-mat[2], mat[0])
        rot_z = 0.0
    rot = (rot_x, rot_y, rot_z)
    return rot

def matrixToYXZ(mat):
    rot_x = 0.0
    rot_y = 0.0
    rot_z = 0.0
    rot_x = -math.asin(mat[5])
    if (math.cos(rot_x) != 0.0):
        rot_z = math.atan2(mat[3], mat[4])
        alpha = mat[2]/math.cos(rot_x)
        if alpha > 1.0:
            alpha = 1.0
        if alpha < -1.0:
            alpha = -1.0
        rot_y = math.asin(alpha)
        if (mat[8] < 0.0):
            rot_y = math.pi - rot_y
    else:
        rot_z = math.atan2(mat[1], mat[0])
        rot_y = 0.0
    rot = (rot_x, rot_y, rot_z)
    return rot

def matrixToXYZ(mat):
    rot_x = 0.0
    rot_y = 0.0
    rot_z = 0.0
    rot_y = -math.asin(mat[2])
    if (math.cos(rot_y) != 0.0):
        rot_z = math.atan2(mat[1], mat[0])
        alpha = mat[5]/math.cos(rot_y)
        if alpha > 1.0:
            alpha = 1.0
        if alpha < -1.0:
            alpha = -1.0
        rot_x = math.asin(alpha)
        if (mat[8] < 0.0):
            rot_x = math.pi - rot_x
    else:
        rot_z = math.atan2(-mat[3], mat[4])
        rot_x = 0.0
    rot = (rot_x, rot_y, rot_z)
    return rot

def matrixToZYX(mat):
    rot_x = 0.0
    rot_y = 0.0
    rot_z = 0.0
    rot_y = -math.asin(mat[6])
    if (math.cos(rot_y) != 0.0):
        rot_x = math.atan2(mat[7], mat[8])
        alpha = mat[3]/math.cos(rot_y)
        if alpha > 1.0:
            alpha = 1.0
        if alpha < -1.0:
            alpha = -1.0
        rot_z = math.asin(alpha)
        if (mat[0] < 0.0):
            rot_z = math.pi - rot_z
    else:
        rot_x = math.atan2(-mat[5], mat[4])
        rot_z = 0.0
    rot = (rot_x, rot_y, rot_z)
    return rot


# class
# --- bone data struct
#
class BoneProperty:
    def __init__(self, frame, name, pos, rot, adj):
        self.frame = frame
        self.name = name
        self.position = pos
        self.rotate = rot
        self.adjust = adj

    def getFrame(self):
        return self.frame

    def getName(self):
        return self.name

    def getPosition(self):
        return self.position

    def getRotate(self, rotzero):
        return self.quaternion2rotation(self.rotate, rotzero)

    def quaternion2rotation(self, qrot, rotzero):
        x = -qrot[0]
        y = -qrot[1]
        z = -qrot[2]
        w = qrot[3]
        mat1 = []
        mat1.append(1-2*y*y-2*z*z)
        mat1.append(2*x*y+2*w*z)
        mat1.append(2*x*z-2*w*y)
        mat1.append(2*x*y-2*w*z)
        mat1.append(1-2*x*x-2*z*z)
        mat1.append(2*y*z+2*w*x)
        mat1.append(2*x*z+2*w*y)
        mat1.append(2*y*z-2*w*x)
        mat1.append(1-2*x*x-2*y*y)
        # RotZero * rotation * RevRotZero
        mat = [1,0,0,0,1,0,0,0,1]
        mat = matmul(mat, matroty(rotzero[2]))
        mat = matmul(mat, matrotx(rotzero[0]))
        mat = matmul(mat, matrotz(rotzero[1]))
        revmat = matrevrot(mat)
        mat2 = matmul(revmat, mat1)
        mat = matmul(mat2, mat)
        # matrix to ZXY type
        rot_xyz = matrixToZXY(mat)
        # convert to degree
        deg = 360.0 / (math.pi * 2.0)
        rot_x = (rot_xyz[0] * deg)
        rot_y = (rot_xyz[1] * deg)
        rot_z = (rot_xyz[2] * deg)
        # result
        v_rot = [rot_x, rot_y, rot_z]
        return v_rot
        

# class
# --- modo bone item struct
#
class ModoItemProperty:
    def __init__(self, name):
        self.name = name
        self.rot = [0.0, 0.0, 0.0]

    def queryRotationZero(self):
        mat = [1, 0, 0, 0, 1, 0, 0, 0, 1]
        name = self.name
        while name != None:
            index = 0
            count = 0
            list = lx.eval('query sceneservice item.xfrmItems ? "%s"' % name)
            # search rotate zero
            for i in list:
                if re.match('^rotation', i) != None:
                    count += 1
                    if count == 2:
                        break
                index += 1
            # get channel
            lx.eval('query sceneservice item.name ? "%s"' % list[index])
            lx.eval('select.channel {%s:rot.X}' % list[index])
            rot_x = lx.eval('channel.value ? channel:{%s:rot.X}' % list[index])
            rot_y = lx.eval('channel.value ? channel:{%s:rot.Y}' % list[index])
            rot_z = lx.eval('channel.value ? channel:{%s:rot.Z}' % list[index])
            # to matrix
            lmat = [1, 0, 0, 0, 1, 0, 0, 0, 1]
            lmat = matmul(lmat, matroty(rot_y))
            lmat = matmul(lmat, matrotx(rot_x))
            lmat = matmul(lmat, matrotz(rot_z))
            mat = matmul(lmat, mat)
            # next parent
            parent = lx.eval('query sceneservice item.parent ? "%s"' % name)
            if parent == None:
                break
            if lx.eval('query sceneservice item.type ? "%s"' % parent) != 'locator':
                break
            name = lx.eval('query sceneservice item.name ? "%s"' % parent)
        # matrix to ZXY type
        rot_xyz = matrixToYXZ(mat)
        self.rot[0] = rot_xyz[0]
        self.rot[1] = rot_xyz[1]
        self.rot[2] = rot_xyz[2]

    def getName(self):
        sname = self.name
        return sname.decode('shift_jis','ignore')

    def getRotationZero(self):
        return self.rot


# class
# --- VMD read precedure
#
class VmdReader:
    def __init__(self):
        self.bone = []
        self.last = 0

    def size(self):
        return len(self.bone)

    def getBone(self, index):
        if (index < 0) or (index > len(self.bone)):
            return None
        return self.bone[index]

    def getLastFrame(self):
        return self.last

    def load(self, fp):
        self.file = fp
        # - header
        h_magic = ReadStringSjis(fp, 30)
        h_name = ReadStringSjis(fp, 20)
        # - header check
        if h_magic != 'Vocaloid Motion Data 0002':
            return
        # - bone
        bone_num = ReadStruct(fp, 'L')
        for loop in range(bone_num):
            b_name = ReadStringSjis(fp, 15)
            b_frame = ReadStruct(fp, 'L')
            b_pos_x = ReadStruct(fp, 'f')
            b_pos_y = ReadStruct(fp, 'f')
            b_pos_z = ReadStruct(fp, 'f')
            b_rot_x = ReadStruct(fp, 'f')
            b_rot_y = ReadStruct(fp, 'f')
            b_rot_z = ReadStruct(fp, 'f')
            b_rot_w = ReadStruct(fp, 'f')
            b_adjust = ReadStruct(fp, '64B')
            v_pos = [b_pos_x, b_pos_y, b_pos_z]
            v_rot = [b_rot_x, b_rot_y, b_rot_z, b_rot_w]
            self.bone.append(BoneProperty(b_frame,b_name,v_pos,v_rot,b_adjust))
            if self.last < b_frame:
                self.last = b_frame
        # - morph
        morph_num = ReadStruct(fp, 'L')
        for loop in range(morph_num):
            m_name =  ReadStringSjis(fp, 15)
            m_frame = ReadStruct(fp, 'L')
            m_value = ReadStruct(fp, 'f')
            if self.last < m_frame:
                self.last = m_frame
            # drop
        # - camera
        camera_num = ReadStruct(fp, 'L')
        for loop in range(camera_num):
            cam_frame = ReadStruct(fp, 'L')
            cam_dist = ReadStruct(fp, 'f')
            cam_pos_x = ReadStruct(fp, 'f')
            cam_pos_y = ReadStruct(fp, 'f')
            cam_pos_z = ReadStruct(fp, 'f')
            cam_rot_x = ReadStruct(fp, 'f')
            cam_rot_y = ReadStruct(fp, 'f')
            cam_rot_z = ReadStruct(fp, 'f')
            cam_adjust = ReadStruct(fp, '24B')
            cam_viewangle = ReadStruct(fp, 'L')
            cam_parth = ReadStruct(fp, 'B')
            if self.last < cam_frame:
                self.last = cam_frame
            # drop
        # - light
        light_num = ReadStruct(fp, 'L')
        for loop in range(camera_num):
            l_frame = ReadStruct(fp, 'L')
            l_color_r = ReadStruct(fp, 'f')
            l_color_g = ReadStruct(fp, 'f')
            l_color_b = ReadStruct(fp, 'f')
            l_pos_x = ReadStruct(fp, 'f')
            l_pos_y = ReadStruct(fp, 'f')
            l_pos_z = ReadStruct(fp, 'f')
            if self.last < l_frame:
                self.last = l_frame
            # drop
        # end
        

#
# main
#

# --- file open dialog
lx.eval('dialog.setup fileOpen')
lx.eval('dialog.title "Open VMD file."')
lx.eval('dialog.msg "select VMD file"')
lx.eval('dialog.open')
filename = lx.eval('dialog.result ?')
# --- load VMD file 
vmd = VmdReader()
with open(filename, "rb") as fp:
    vmd.load(fp)
    bone_num = vmd.size()
    if bone_num == 0:
        # error
        lx.eval('dialog.setup error')
        lx.eval('dialog.title "VMD Open error."')
        lx.eval('dialog.msg "file cannot read or empty file."')
        lx.eval('dialog.open')
    else:
        # -- modo bone item list
        locator_list = []
        itemnum = lx.eval('query sceneservice item.N ? all')
        for i in range(itemnum):
            if lx.eval("query sceneservice item.type ? %d" % i)=="locator":
                itemname = lx.eval("query sceneservice item.name ? %d" % i)
		locator_list.append(ModoItemProperty(itemname))
        for j in locator_list:
            j.queryRotationZero()
        # -- setup to modo
        # anime frame
        #lx.eval("time.fpsCustom 30.0")
        lx.eval("time.fps ntsc30")
        lastframe = vmd.getLastFrame() / 30.0
        lx.eval("time.range scene out:%f" % lastframe)
        lx.eval("time.range current out:%f" % lastframe)
        lx.eval("select.time 0.0 0")
        oneframe = 1.0 / 30.0
        # position magnify
        lx.eval('user.defNew pos_mag float momentary')
        lx.eval('user.value pos_mag 1.0')
        lx.eval('user.def pos_mag username "Position magnification"')
        lx.eval('user.value pos_mag')
        posmag = lx.eval('user.value pos_mag ?')
        # bone set
        for index in range(bone_num):
            bone = vmd.getBone(index)
            for itemname in locator_list:
                if itemname.getName() == bone.getName():
                    # select item
                    lx.eval('select.subItem "%s" set locator' % bone.getName().encode('shift_jis'))
                    # frame jump
                    frame = bone.getFrame() * oneframe
                    lx.eval("select.time %f 0" % frame)
                    # setting rotation
                    rot = bone.getRotate(itemname.getRotationZero())
                    lx.eval("transform.channel rot.X %f" % rot[0])
                    lx.eval("transform.channel rot.Y %f" % rot[1])
                    lx.eval("transform.channel rot.Z %f" % rot[2])
                    lx.eval('transform.key rot')
                    # setting position
                    pos = bone.getPosition()
                    if (pos[0] != 0.0) or (pos[1] != 0.0) or (pos[2] != 0.0):
                        lx.eval("transform.channel pos.X %f" % (pos[0]*posmag))
                        lx.eval("transform.channel pos.Y %f" % (pos[1]*posmag))
                        lx.eval("transform.channel pos.Z %f" % (pos[2]*posmag))
                        lx.eval('transform.key pos')
                    # debug
                    #lx.out("set bone name %s %d %d" % (bone.getName().encode('shift_jis') , bone.getFrame(), index))
                    

